import dataclasses
import pathlib
from collections import OrderedDict
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
    cast,
)

from .types import MAPPING_TYPES, PRIMITIVE_TYPES, SEQUENCE_TYPES

TypeHookCallback = Callable[[Any], Any]


def to_dash_case(s: str) -> str:
    # convert field name to dash-case
    # Example: snake_case to snake-case
    return s.replace("_", "-")


@dataclasses.dataclass
class SettingBase:
    def asdict(
        self,
        ignore_fields: Optional[Sequence[str]] = None,
        omit_none: bool = False,
        naming_rule: Optional[Callable[[str], str]] = None,
        renames: Optional[Dict[str, str]] = None,
        type_hooks: Optional[Dict[Type[Any], TypeHookCallback]] = None,
    ) -> Dict[str, Any]:
        entries = dataclasses.asdict(self)
        if ignore_fields is not None:
            for ignore in ignore_fields:
                assert ignore in entries
                entries.pop(ignore)

        if omit_none:
            empty = [k for k, v in entries.items() if v is None]
            for k in empty:
                entries.pop(k)

        if renames is not None:
            for src, to in renames.items():
                if src in entries:
                    value = entries.pop(src)
                    entries[to] = value

        field_name_converter: Callable[[str], str] = lambda s: s
        if naming_rule is not None:
            field_name_converter = naming_rule

        type_conversion: Dict[Type[Any], TypeHookCallback] = {
            pathlib.Path: lambda x: str(x),
        }
        if type_hooks is not None:
            type_conversion.update(type_hooks)

        def convert_types(data: Any) -> Any:
            if data is None:
                return None

            data_type = type(data)
            conversion = type_conversion.get(data_type, None)
            if conversion is not None:
                return conversion(data)

            if isinstance(data, PRIMITIVE_TYPES):
                return data
            elif isinstance(data, SEQUENCE_TYPES):
                return list(sorted([convert_types(x) for x in data]))
            elif isinstance(data, MAPPING_TYPES):
                # NOTE(igarashi): key must be a str
                return OrderedDict(
                    (field_name_converter(k), convert_types(v))
                    for k, v in sorted(data.items())
                )
            else:
                raise RuntimeError(
                    f"cannot handle type: {data_type}, consider adding type_hooks"
                )

        return cast(Dict[str, Any], convert_types(entries))


SectionPathType = Sequence[str]
SectionDataType = Dict[str, Any]


def _sort_object(data: Any) -> Any:
    if data is None:
        return None

    if isinstance(data, PRIMITIVE_TYPES):
        return data
    elif isinstance(data, SEQUENCE_TYPES):
        return list(sorted(data))
    elif isinstance(data, MAPPING_TYPES):
        return OrderedDict((k, _sort_object(v)) for k, v in sorted(data.items()))
    else:
        raise RuntimeError(f"cannot handle type: {type(data)}")


def _create_dict(paths: SectionPathType) -> Any:
    element: Dict[str, Any] = {}
    for path in paths[::-1]:
        element = {path: element}
    return element


def _traverse_toml(
    section_path: SectionPathType, toml: Dict[str, Any], create: bool
) -> Any:
    element = toml
    current_path: List[str] = []
    for idx, path in enumerate(section_path):
        if path in element:
            element = element[path]
        elif create:
            # create remaining items and set create=False so that we don't create nodes anymore
            element[path] = _create_dict(section_path[idx + 1 :])
            create = False

            # re-create the reference from toml root
            element = _traverse_toml(section_path[: idx + 1], toml, False)
        else:
            raise ValueError(f"Key {path} does not exist")

        current_path.append(path)
        if not isinstance(element, dict):
            raise KeyError(f"invalid section, {current_path} exists")

    return element


class SettingFile:
    def __init__(self) -> None:
        self._entries: List[Tuple[SectionPathType, SectionDataType]] = []
        self._structure: Set[Tuple[str, ...]] = set()

    def entries(self) -> List[Tuple[SectionPathType, Dict[str, Any]]]:
        return self._entries

    @staticmethod
    def update_by_entry(
        dst: Dict[str, Any],
        section_path: SectionPathType,
        section_data: SectionDataType,
        replace: bool = True,
    ) -> Dict[str, Any]:
        data: Dict[str, Any] = _traverse_toml(section_path, dst, True)
        parent = _traverse_toml(section_path[:-1], dst, False)

        if replace:
            # NOTE(igarashi): tomlkit.TOMLDocument doesn't clear its dictionary by `clear()`,
            # so we set a new dict from its parent
            parent[section_path[-1]] = {}
            data = _traverse_toml(section_path, dst, False)

        # NOTE(igarashi): don't use dict.update so that the items are added
        # in alphabetical order.
        for key, value in sorted(section_data.items()):
            data[key] = _sort_object(value)

        return data

    def as_dict(self) -> Dict[str, Any]:
        root: Dict[str, Any] = {}
        for entry in self._entries:
            SettingFile.update_by_entry(root, entry[0], entry[1])

        return root

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, dict):
            return other == self.as_dict()

        return id(other) == id(self)

    def get_section(
        self,
        path: Union[str, SectionPathType],
        default: Optional[SectionDataType] = None,
    ) -> SectionDataType:
        target: Tuple[str, ...]
        if isinstance(path, str):
            target = (path,)
        else:
            target = tuple(path)

        for p, d in self._entries:
            if p == target:
                return d

        if default is not None:
            return default

        raise KeyError(f"section: {path} not found")

    def set_section(self, path: SectionPathType, data: SectionDataType) -> None:
        assert isinstance(path, (tuple, list))
        subpath: List[str] = []
        for p in path:
            if tuple(subpath) in self._structure:
                raise KeyError(f"subpath: {subpath} already exists")

            subpath.append(p)

        self._entries.append((tuple(path), data))



import logging
import re
from pathlib import Path
from typing import Callable, Generator, Iterable, Optional

import unidiff

from pysen.diagnostic import Diagnostic
from pysen.exceptions import UnexpectedErrorFormat

FilePathParserType = Callable[[str], Path]
_logger = logging.getLogger(__name__)


def _warn_parse_error(errors: str, logger: Optional[logging.Logger]) -> None:
    logger = logger or _logger
    logger.warning(
        "The following error(s) could not be parsed, so pysen won't format it (them).\n"
        f"{errors}\n"
        f"If you think this is a bug please report it to the maintainers."
    )


def parse_error_lines(
    errors: str, logger: Optional[logging.Logger] = None
) -> Iterable[Diagnostic]:
    """
    Compatible with flake8, mypy
    """
    number = r"(?:0|[1-9]\d*)"
    _file_path = r"^(?P<file_path>.*?)"
    _line = rf":(?P<line>{number})"
    _column = rf"(:(?P<column>{number}))?"
    _message = r": (?P<message>.*$)"
    pattern = _file_path + _line + _column + _message
    invalid_lines = []
    for el in errors.splitlines():
        m = re.match(pattern, el)
        if m is None:
            invalid_lines.append(el)
            continue
        line = int(m.group("line"))
        if m.group("column") is None:
            column = None
        else:
            column = int(m.group("column"))
        yield Diagnostic(
            start_line=line,
            end_line=line,
            start_column=column,
            message=m.group("message").lstrip(" ").rstrip("\n"),
            file_path=Path(m.group("file_path")),
        )
    if invalid_lines:
        _warn_parse_error("\n".join(invalid_lines), logger)


def parse_error_diffs(
    errors: str,
    file_path_parser: FilePathParserType,
    logger: Optional[logging.Logger] = None,
) -> Iterable[Diagnostic]:
    """
    Compatible with isort, black
    """

    def _is_changed(line: unidiff.patch.Line) -> bool:
        return not line.is_context

    try:
        patches = unidiff.PatchSet(errors)
    except unidiff.errors.UnidiffParseError:
        _warn_parse_error(errors, logger)
        return
    for patch in patches:
        for hunk in patch:
            source_changes = list(filter(_is_changed, hunk.source_lines()))
            if source_changes:
                start_line = source_changes[0].source_line_no
                end_line = source_changes[-1].source_line_no
            else:
                target_changes = list(filter(_is_changed, hunk.target_lines()))
                assert target_changes, "expected either source or target line number"
                start_line = target_changes[0].target_line_no
                end_line = target_changes[-1].target_line_no

            try:
                file_path = file_path_parser(patch.source_file)
            except UnexpectedErrorFormat:
                _warn_parse_error(patch, logger)
                continue

            def filter_hunk(
                hunk: unidiff.patch.Hunk,
            ) -> Generator[unidiff.patch.Line, None, None]:
                for line in hunk:
                    if _is_changed(line):
                        yield line
                    elif line.source_line_no is not None:
                        if start_line <= line.source_line_no <= end_line:
                            yield line

            yield Diagnostic(
                start_line=start_line,
                end_line=end_line,
                start_column=1,
                file_path=file_path,
                diff="".join(map(str, filter_hunk(hunk))),
            )



import copy
import dataclasses
import enum
import functools
import pathlib
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple

from pysen import process_utils
from pysen.command import check_command_installed
from pysen.dist_version import get_version
from pysen.error_lines import parse_error_diffs
from pysen.exceptions import IncompatibleVersionError, UnexpectedErrorFormat
from pysen.path import change_dir
from pysen.py_version import VersionRepresentation
from pysen.reporter import Reporter
from pysen.setting import SettingBase

_SettingFileName = "pyproject.toml"


class IsortSectionName(enum.Enum):
    FUTURE = "FUTURE"
    STDLIB = "STDLIB"
    THIRDPARTY = "THIRDPARTY"
    FIRSTPARTY = "FIRSTPARTY"
    LOCALFOLDER = "LOCALFOLDER"


@functools.lru_cache(1)
def _get_isort_version() -> VersionRepresentation:
    version = get_version("isort")
    if version.major not in [4, 5]:
        raise IncompatibleVersionError(
            "pysen only supports isort versions 4 and 5. "
            f"version {version} is not supported."
        )

    return version


def _check_version_compatibility(
    ensure_newline_before_comments: Optional[bool],
    version: VersionRepresentation,
) -> None:
    if version.major == 4 and ensure_newline_before_comments is not None:
        raise IncompatibleVersionError(
            "isort option `ensure_newline_before_comments`"
            f"is not supported in your isort version {version}"
        )


@dataclasses.dataclass
class IsortSetting(SettingBase):
    force_grid_wrap: int = 0
    force_single_line: bool = False
    include_trailing_comma: bool = True
    known_first_party: Optional[Set[str]] = None
    known_third_party: Optional[Set[str]] = None
    line_length: int = 88
    multi_line_output: int = 3
    default_section: Optional[IsortSectionName] = None
    sections: Optional[List[IsortSectionName]] = None
    use_parentheses: bool = True
    ensure_newline_before_comments: Optional[bool] = None

    @staticmethod
    def default() -> "IsortSetting":
        return IsortSetting()

    def to_black_compatible(self) -> "IsortSetting":
        # NOTE(igarashi)
        # multi_line_output: black uses 3 (Vertical Hanging Indent)
        # include_trailing_comma: black appends trailing comma
        # force_grid_wrap: the property means isort grid-wrap the statement regardless
        #                  of line length if the number of `from` imports is greater than
        #                  the property. black doesn't grid wrap the statement if it
        #                  doesn't exceed the line length.
        # use_parentheses: use parenthesis for line continuation instead of `\`
        new = copy.deepcopy(self)

        new.multi_line_output = 3
        new.include_trailing_comma = True
        new.force_grid_wrap = 0
        new.use_parentheses = True

        # See issue #277
        isort_version = _get_isort_version().major
        if isort_version >= 5:
            new.ensure_newline_before_comments = True
        return new

    def export(self) -> Tuple[List[str], Dict[str, Any]]:
        section_name = ["tool", "isort"]

        _check_version_compatibility(
            self.ensure_newline_before_comments,
            _get_isort_version(),
        )
        entries = self.asdict(
            omit_none=True, type_hooks={IsortSectionName: lambda x: x.value}
        )
        return section_name, entries


def _parse_file_path(file_path: str) -> pathlib.Path:
    ret = file_path.split(" ")[0]
    before_suffix = ":before"
    after_suffix = ":after"
    if ret.endswith(before_suffix):
        return pathlib.Path(ret.rsplit(before_suffix, 1)[0])
    elif ret.endswith(after_suffix):
        return pathlib.Path(ret.rsplit(after_suffix, 1)[0])
    else:
        raise UnexpectedErrorFormat(file_path)


def run(
    reporter: Reporter,
    base_dir: pathlib.Path,
    setting_path: pathlib.Path,
    sources: Iterable[pathlib.Path],
    inplace_edit: bool,
) -> int:
    check_command_installed(*process_utils.add_python_executable("isort", "--version"))
    version = _get_isort_version()

    targets = [str(d) for d in sources]
    if len(targets) == 0:
        return 0

    cmd = ["isort", "--settings-path", str(setting_path)]
    if version.major == 4:
        cmd.append("--recursive")
    if not inplace_edit:
        cmd += ["--diff", "--check-only"]
    cmd += targets

    with change_dir(base_dir):
        ret, stdout, _ = process_utils.run(
            process_utils.add_python_executable(*cmd), reporter
        )

    diagnostics = parse_error_diffs(stdout, _parse_file_path, logger=reporter.logger)
    reporter.report_diagnostics(list(diagnostics))

    return ret



import contextlib
import io
import logging
import threading
import time
from typing import Any, Iterator, List, Optional, Sequence

from .diagnostic import Diagnostic, DiagnosticFormatter
from .logging_utils import (
    CommandLoggingOptions,
    get_process_output_logger,
    get_reporter_logger,
)

_COMMAND_REPR_MAX_LENGTH = 150
_OMIT_REPR = "..."


def _truncate_command_sequence(cmds: str) -> str:
    if len(cmds) <= _COMMAND_REPR_MAX_LENGTH:
        return cmds

    prefix_length = _COMMAND_REPR_MAX_LENGTH - len(_OMIT_REPR)
    assert prefix_length > 0
    prefix = cmds[:prefix_length]
    return f"{prefix}{_OMIT_REPR}"


class Reporter:
    def __init__(self, name: str) -> None:
        self._name = name
        self._success: Optional[bool] = None
        self._exit_code: Optional[int] = None

        self._commands: List[str] = []
        self._diagnostics: List[Diagnostic] = []
        self._started: Optional[float] = None
        self._ended: Optional[float] = None

        self._logger = get_reporter_logger(name)
        self._process_output = get_process_output_logger(name)

    @property
    def name(self) -> str:
        return self._name

    @property
    def success(self) -> bool:
        assert self._success is not None
        return self._success

    @property
    def exit_code(self) -> int:
        assert self._exit_code is not None
        return self._exit_code

    @property
    def commands(self) -> List[str]:
        assert self._commands is not None
        return self._commands

    @property
    def elapsed_time(self) -> float:
        assert self._started is not None
        assert self._ended is not None
        return self._ended - self._started

    @property
    def diagnostics(self) -> List[Diagnostic]:
        return self._diagnostics

    def __enter__(self) -> "Reporter":
        self._started = time.time()
        self.logger.info(f"Running: {self.name}")
        return self

    def __exit__(self, exc_type: Any, exc: Any, tb: Any) -> None:
        self._ended = time.time()

    def report_diagnostics(self, diagnostics: Sequence[Diagnostic]) -> None:
        self._diagnostics.extend(diagnostics)

    def report_command(self, cmd: str) -> None:
        self._logger.debug(f"> {_truncate_command_sequence(cmd)}")
        self._commands.append(cmd)

    @property
    def logger(self) -> logging.Logger:
        return self._logger

    @property
    def process_output(self) -> logging.Logger:
        return self._process_output

    def set_result(self, success: bool, exit_code: Optional[int] = None) -> None:
        self._success = success
        self._exit_code = exit_code


class ReporterFactory:
    def __init__(
        self,
        pretty: bool = True,
        process_output: bool = True,
        loglevel: int = logging.INFO,
    ) -> None:
        self._reporters: List[Reporter] = []
        self._lock = threading.Lock()
        self._pretty = pretty
        self._process_output = process_output
        self._loglevel = loglevel

    def create(self, name: str) -> Reporter:
        r = Reporter(name)
        with self._lock:
            self._reporters.append(r)
        return r

    @contextlib.contextmanager
    def logging_handlers(self, is_grouped: bool) -> Iterator[None]:
        clo = CommandLoggingOptions(is_grouped, self._pretty, self._process_output)
        with clo.start_logging(self._loglevel):
            yield

    def has_error(self) -> bool:
        return not all([r.success for r in self._reporters])

    @property
    def reporters(self) -> List[Reporter]:
        return self._reporters

    def format_summary(self) -> str:
        with io.StringIO() as buf:
            for r in self._reporters:
                status_msg = "Failed"
                if r.success:
                    status_msg = "OK"

                buf.write(
                    "{} .......... {} ({:.2f} sec)\n".format(
                        r.name, status_msg, r.elapsed_time
                    )
                )

            return buf.getvalue()

    def format_error_summary(self) -> str:
        with io.StringIO() as buf:
            buf.write("Error:\n")
            for r in self._reporters:
                if not r.success:
                    buf.write(" - {}\n".format(r.name))
            return buf.getvalue()

    def format_diagnostic_summary(self, formatter: DiagnosticFormatter) -> str:
        ret: List[str] = []
        for r in self._reporters:
            for d in r.diagnostics:
                ret.append(formatter.format(d, r.name))

        return "\n".join(ret)



import argparse
import concurrent.futures
import contextlib
import logging
import pathlib
import tempfile
from typing import Dict, List, Optional, Sequence

from . import path
from .command import CommandBase
from .exceptions import (
    CommandNotFoundError,
    InvalidCommandNameError,
    RunTargetFileNotSupported,
)
from .manifest import ManifestBase, ParserType, TargetType
from .reporter import ReporterFactory
from .runner_options import PathContext, RunOptions
from .types import ComponentName, TargetName

_logger = logging.getLogger(__name__)


def _verify_command_name(command: CommandBase) -> None:
    if ":" in command.name:
        raise InvalidCommandNameError(command.name)


def _has_side_effects(target: TargetType) -> bool:
    return any(cmd.has_side_effects for cmd in target)


def run_target(
    target: TargetType,
    reporters: ReporterFactory,
    options: RunOptions,
    files: Optional[Sequence[pathlib.Path]] = None,
) -> None:
    def run_cmd(cmd: CommandBase) -> bool:
        _verify_command_name(cmd)
        with reporters.create(cmd.name) as r:
            exit_code: int
            try:
                if files is not None:
                    exit_code = cmd.run_files(reporter=r, files=files)
                else:
                    exit_code = cmd.run(reporter=r)
            except CommandNotFoundError:
                exit_code = 127
                r.logger.exception("command not found")
            except RunTargetFileNotSupported:
                exit_code = 0
                r.logger.info(f"{cmd.name} does not support target file execution")
            except KeyboardInterrupt:
                exit_code = 130
                r.logger.exception("interrupted")
                return False
            except BaseException:
                exit_code = -1
                r.logger.exception("unexpected exception")
            finally:
                r.set_result(exit_code == 0, exit_code)
        return True

    if options.no_parallel:
        is_grouped = False
    else:
        is_grouped = not _has_side_effects(target)

    with reporters.logging_handlers(is_grouped=is_grouped):
        if is_grouped:
            # TODO: control the maximum number of concurrent threads
            _logger.info("Running commands concurrently...")
            with concurrent.futures.ThreadPoolExecutor() as executor:
                executor.map(run_cmd, target)
            _logger.info("... concurrent execution done")
        else:
            _logger.info("Running commands")
            for cmd in target:
                if not run_cmd(cmd):
                    break


class Runner:
    def __init__(self, manifest: ManifestBase) -> None:
        self._manifest = manifest

    def setup_manifest_argparse(self, parser: ParserType) -> None:
        self._manifest.configure_parser(parser)

    def parse_manifest_arguments(self, args: Sequence[str]) -> argparse.Namespace:
        parser = argparse.ArgumentParser("manifest options")
        self.setup_manifest_argparse(parser)
        return parser.parse_args(args)

    def export_settings(
        self,
        base_dir: pathlib.Path,
        settings_dir: pathlib.Path,
        args: argparse.Namespace,
    ) -> None:
        self._manifest.export_settings(PathContext(base_dir, settings_dir), args)

    def get_targets(self, args: argparse.Namespace) -> Dict[str, List[ComponentName]]:
        return self._manifest.get_targets(args)

    def _get_target(
        self,
        target_name: TargetName,
        paths: PathContext,
        options: RunOptions,
        args: argparse.Namespace,
    ) -> TargetType:
        # NOTE(igarashi): make sure if target_name exists before calling get_target(target_name)
        targets = self.get_targets(args)
        if target_name not in targets:
            raise CommandNotFoundError(f"target: {target_name} not found")

        return self._manifest.get_target(target_name, paths, options, args)

    def run(
        self,
        target_name: str,
        base_dir: pathlib.Path,
        manifest_args: argparse.Namespace,
        reporters: ReporterFactory,
        options: RunOptions,
        settings_dir: Optional[pathlib.Path] = None,
        files: Optional[Sequence[pathlib.Path]] = None,
    ) -> None:
        with contextlib.ExitStack() as stack:
            base_dir = base_dir.resolve()
            if settings_dir is not None:
                settings_dir = path.resolve_path(base_dir, settings_dir)
                settings_dir.mkdir(parents=True, exist_ok=True)
            else:
                tempdir = stack.enter_context(tempfile.TemporaryDirectory())
                settings_dir = pathlib.Path(tempdir)

            paths = PathContext(base_dir, settings_dir)
            self.export_settings(paths.base_dir, paths.settings_dir, manifest_args)
            target = self._get_target(target_name, paths, options, manifest_args)
            run_target(target, reporters, options, files)



import dataclasses
import functools
import pathlib
from typing import Any, Dict, Iterable, List, Optional, Tuple

from pysen import process_utils
from pysen.command import check_command_installed
from pysen.dist_version import get_version
from pysen.error_lines import parse_error_diffs
from pysen.exceptions import IncompatibleVersionError
from pysen.path import change_dir
from pysen.py_version import PythonVersion, VersionRepresentation
from pysen.reporter import Reporter
from pysen.setting import SettingBase, to_dash_case


@dataclasses.dataclass
class BlackSetting(SettingBase):
    line_length: int = 88
    target_version: List[PythonVersion] = dataclasses.field(default_factory=list)

    @staticmethod
    def default(
        # NOTE(igarashi) safe to use as an argument since it is immutable
        py_version: Optional[PythonVersion] = None,
    ) -> "BlackSetting":
        py_version = py_version or PythonVersion(3, 7)
        return BlackSetting(target_version=[py_version])

    def export(self) -> Tuple[List[str], Dict[str, Any]]:
        section_name = ["tool", "black"]

        # TODO(igarashi): refactor these flaky code
        # `dataclass.asdict()` converts `PythonVersion` object into an undesired dict like
        # `{"major": ...}` as it is defined by a dataclass.
        # `SettingBase.asdict` takes `type_hooks` argument, but it doesn't work as expected
        # because the hook is called after `dataclass.asdict()` is called.
        # In order to dump PythonVersion objects into a desired representation,
        # the following code ignores the field when calling `asdict` and replaces it
        # with the desired dump after we call `asdict()`
        entries = self.asdict(
            ignore_fields=["target_version"], naming_rule=to_dash_case
        )
        entries["target-version"] = []
        for v in self.target_version:
            entries["target-version"].append(v.short_representation)
        return section_name, entries


def _parse_file_path(file_path: str) -> pathlib.Path:
    return pathlib.Path(file_path.split(" ")[0])


@functools.lru_cache(1)
def _check_black_version() -> None:
    version = get_version("black")
    minimum_supported = VersionRepresentation(19, 10)

    if version < minimum_supported:
        raise IncompatibleVersionError(
            f"pysen only supports black >= {minimum_supported}, "
            f"version {version} is not supported."
        )


def run(
    reporter: Reporter,
    base_dir: pathlib.Path,
    setting_path: pathlib.Path,
    sources: Iterable[pathlib.Path],
    inplace_edit: bool,
) -> int:
    check_command_installed(*process_utils.add_python_executable("black", "--version"))
    _check_black_version()

    targets = [str(d) for d in sources]
    if len(targets) == 0:
        return 0

    cmd = (
        ["black", "--config", str(setting_path)]
        + (["--diff", "--check"] if not inplace_edit else [])
        + targets
    )
    with change_dir(base_dir):
        ret, stdout, _ = process_utils.run(
            process_utils.add_python_executable(*cmd), reporter
        )

    diagnostics = parse_error_diffs(stdout, _parse_file_path, logger=reporter.logger)
    reporter.report_diagnostics(list(diagnostics))

    return ret



import pathlib
from abc import ABC, abstractmethod
from typing import Sequence

from .component import ComponentBase
from .pyproject_model import Config, PluginConfig


class PluginBase(ABC):
    @abstractmethod
    def load(
        self, file_path: pathlib.Path, config: PluginConfig, root: Config
    ) -> Sequence[ComponentBase]:
        ...



import importlib
import pathlib
from types import ModuleType
from typing import Optional

from . import py_module
from .exceptions import InvalidPluginError
from .plugin import PluginBase
from .py_module import _parse_entry_point


def _load(module: ModuleType, func_name: str) -> PluginBase:
    entry_point = getattr(module, func_name, None)
    if entry_point is None or not callable(entry_point):
        raise InvalidPluginError(
            module.__file__, f"expected to have `{func_name}` method: {module.__file__}"
        )

    try:
        ret = entry_point()
    except Exception as e:
        raise RuntimeError(
            f"an error occurred while loading {module.__file__}::{func_name}",
        ) from e

    if not isinstance(ret, PluginBase):
        raise InvalidPluginError(
            module.__file__, f"`{func_name}` must return an instance of PluginBase"
        )

    return ret


def load_from_file(path: pathlib.Path) -> PluginBase:
    module = py_module.load(path, "plugin")
    return _load(module, "plugin")


def load_from_module(entry_point: str) -> PluginBase:
    parsed = _parse_entry_point(entry_point)
    if parsed is None:
        raise ValueError(f"invalid entry_point: {entry_point}")

    module_name, func_name = parsed
    module = importlib.import_module(module_name)
    return _load(module, func_name)


def load_plugin(
    function: Optional[str] = None, script: Optional[pathlib.Path] = None
) -> PluginBase:
    if function is None and script is None:
        raise TypeError("must specify either function or script")

    if function is not None and script is not None:
        raise TypeError("only one of function or script must be specified")

    if function is not None:
        try:
            return load_from_module(function)
        except BaseException:
            if script is not None:
                pass  # fallback
            else:
                raise

    assert script is not None
    return load_from_file(script)



from abc import ABC, abstractmethod
from typing import DefaultDict, Optional, Sequence

from .command import CommandBase
from .runner_options import PathContext, RunOptions
from .setting import SettingFile
from .source import Source
from .types import ComponentName, TargetName


class ComponentBase(ABC):
    @property
    def name(self) -> Optional[ComponentName]:
        return None

    def export_settings(
        self,
        paths: PathContext,
        files: DefaultDict[str, SettingFile],
    ) -> None:
        pass

    @property
    @abstractmethod
    def targets(self) -> Sequence[TargetName]:
        ...

    @abstractmethod
    def create_command(
        self, target: str, paths: PathContext, options: RunOptions
    ) -> CommandBase:
        ...


class LintComponentBase(ComponentBase):
    def __init__(self, name: str, source: Optional[Source] = None) -> None:
        self._name = name
        self._source: Source = source or Source(includes=["."])

    @property
    def name(self) -> ComponentName:
        return self._name

    @property
    def source(self) -> Source:
        return self._source

    @source.setter
    def source(self, value: Source) -> None:
        self._source = value



import logging
from typing import Optional

import pkg_resources

from pysen.exceptions import DistributionNotFound
from pysen.py_version import VersionRepresentation

_logger = logging.getLogger(__name__)


def _get_distro(name: str) -> Optional[pkg_resources.Distribution]:
    try:
        return pkg_resources.get_distribution(name)
    except pkg_resources.DistributionNotFound:
        _logger.debug(f"distribution {name} not found", exc_info=True)
        return None


def get_version(name: str) -> VersionRepresentation:
    distro = _get_distro(name)
    if distro is None:
        raise DistributionNotFound(
            f"Expected {name} to be installed but pkg_resources could not find it.\n"
            f'Hint: Did you install "{name}" in the same Python environment as pysen?'
        )
    return VersionRepresentation.from_str(distro.version)



import argparse
import distutils.core
import functools
import pathlib
import sys
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Type, Union

import setuptools

from .manifest import ManifestBase
from .path import PathLikeType, wrap_path
from .pyproject import find_pyproject, load_manifest
from .reporter import ReporterFactory
from .runner import Runner
from .runner_options import RunOptions

ManifestLikeType = Union[str, pathlib.Path, ManifestBase]
CommandClassType = Type[distutils.core.Command]
UNDEFINED = object()

_PREDEFINED_COMMAND_NAMES = [
    "build_ext",
    "build_py",
    "develop",
    "install",
    "test",
]


def _get_setuptool_command(name: str) -> CommandClassType:
    if name in _PREDEFINED_COMMAND_NAMES:
        try:
            import importlib

            module = importlib.import_module("setuptools.command.{}".format(name))
            klass = getattr(module, name)
            if (
                klass is not None
                and isinstance(klass, type)
                and issubclass(klass, setuptools.Command)
            ):
                return klass
        except BaseException:
            pass  # failover

    return setuptools.Command


def _get_setuptool_user_options(
    klass: CommandClassType,
) -> List[Tuple[str, Optional[str], str]]:
    try:
        base_user_options = getattr(klass, "user_options", None)
        if base_user_options is not None and isinstance(
            base_user_options, (list, tuple)
        ):
            return list(base_user_options)
    except BaseException:
        pass  # failover

    return []


def _create_setuptool_command(
    name: str,
    runner: Runner,
    base_dir: pathlib.Path,
    settings_dir: Optional[pathlib.Path],
    manifest_args: argparse.Namespace,
) -> CommandClassType:
    base_class = _get_setuptool_command(name)
    base_user_options = _get_setuptool_user_options(base_class)

    setup_options = base_user_options

    class Cmd(base_class):  # type: ignore[valid-type,misc]  # NOQA: F821
        user_options = setup_options

        def __invoke_super(self, func: Callable[[], None]) -> None:
            if base_class is not setuptools.Command:
                func()

        def initialize_options(self) -> None:
            self.__invoke_super(super().initialize_options)

        def finalize_options(self) -> None:
            self.__invoke_super(super().finalize_options)

        def run(self) -> None:
            reporters = ReporterFactory()
            options = RunOptions()
            runner.run(
                name,
                base_dir,
                manifest_args,
                reporters,
                options,
                settings_dir=settings_dir,
                files=None,
            )
            print("\n ** execution summary **")
            print(reporters.format_summary())
            if reporters.has_error():
                sys.stderr.write(f"{name} finished with error(s)\n")
                print(reporters.format_error_summary())
                sys.exit(1)

            self.__invoke_super(super().run)

    Cmd.__name__ = name
    return Cmd


class SetupPyWrapper:
    def __init__(self, cmds: Dict[str, CommandClassType]) -> None:
        self._cmds = cmds

    @property
    def cmdclass(self) -> Dict[str, CommandClassType]:
        return self._cmds

    @functools.wraps(setuptools.setup)
    def __call__(self, **kwargs: Any) -> None:
        cmdclass = kwargs.pop("cmdclass", {})
        assert isinstance(cmdclass, dict)
        configured_cmds = self.cmdclass
        # NOTE(igarashi): raise Exception if the key is duplicated
        for key in cmdclass.keys():
            if key in configured_cmds:
                raise RuntimeError(f"cmdclass: {key} is duplicated")

        cmdclass.update(configured_cmds)
        kwargs.update({"cmdclass": cmdclass})

        setuptools.setup(**kwargs)


def _setup(
    package_dir: pathlib.Path,
    settings_dir: Optional[pathlib.Path],
    manifest: ManifestBase,
    args: Sequence[str],
) -> SetupPyWrapper:
    runner = Runner(manifest)
    parsed = runner.parse_manifest_arguments(args)
    targets = runner.get_targets(parsed)
    setup_commands = {
        name: _create_setuptool_command(name, runner, package_dir, settings_dir, parsed)
        for name in targets
    }

    return SetupPyWrapper(setup_commands)


def setup_from_pyproject(
    fpath: str,
    path: Optional[PathLikeType] = None,
    manifest_args: Optional[Sequence[str]] = None,
    settings_dir: Optional[pathlib.Path] = None,
) -> SetupPyWrapper:
    package_dir = pathlib.Path(fpath).resolve().parent
    wrapped: pathlib.Path = package_dir / "pyproject.toml"
    if path is not None:
        wrapped = wrap_path(path)

    args = manifest_args or []
    pyproject = find_pyproject(wrapped)
    manifest = load_manifest(pyproject)

    return _setup(package_dir, settings_dir, manifest, args)


def setup(
    fpath: str,
    manifest: ManifestBase,
    manifest_args: Optional[Sequence[str]] = None,
    settings_dir: Optional[pathlib.Path] = None,
) -> SetupPyWrapper:
    package_dir = pathlib.Path(fpath).resolve().parent
    args = manifest_args or []

    return _setup(package_dir, settings_dir, manifest, args)


def generate_setting_files(
    fpath: str,
    manifest: ManifestBase,
    export_dir: pathlib.Path,
    manifest_args: Optional[Sequence[str]] = None,
) -> None:
    package_dir = pathlib.Path(fpath).resolve().parent
    args = manifest_args or []
    runner = Runner(manifest)
    parsed = runner.parse_manifest_arguments(args)
    runner.export_settings(package_dir, export_dir, parsed)



import argparse
import dataclasses
import enum
import logging
import pathlib
import sys
from typing import Optional, Sequence, Tuple

from . import __version__, cli_config, exceptions
from .cli_config import CliConfig
from .diagnostic import DiagnosticFormatter, FLCMFormatter
from .logging_utils import setup_logger
from .manifest import ManifestBase
from .path import wrap_path
from .pyproject import find_pyproject, load_manifest
from .reporter import ReporterFactory
from .runner import Runner
from .runner_options import RunOptions

CLI_DESCRIPTION = "pysen CLI"


_ErrorFormat = {
    "gnu": FLCMFormatter,
}


@dataclasses.dataclass
class _SetupOptions:
    error_formatter: Optional[DiagnosticFormatter]
    options: RunOptions
    loglevel: int
    process_output: bool


@enum.unique
class LogLevel(enum.Enum):
    debug = logging.DEBUG
    info = logging.INFO
    warning = logging.WARNING
    error = logging.ERROR


@enum.unique
class ProcessOutputMode(enum.Enum):
    auto = enum.auto()
    show = enum.auto()
    hide = enum.auto()


def _get_loglevel(user_specification: Optional[str], default: int) -> int:
    if user_specification is None:
        return default
    else:
        ret: int = LogLevel[user_specification].value
        return ret


def _is_process_output_requested(user_specification: str, default: bool) -> bool:
    user_specification_mode = ProcessOutputMode[user_specification]
    return user_specification_mode == ProcessOutputMode.show or (
        user_specification_mode == ProcessOutputMode.auto and default
    )


def _use_pretty_logging() -> bool:
    return sys.stderr.isatty()


def _show_version() -> None:
    print(__version__)


def _setup_run(
    base_dir: pathlib.Path,
    args: argparse.Namespace,
    config: Optional[CliConfig],
) -> _SetupOptions:
    error_formatter: Optional[DiagnosticFormatter] = None
    default_loglevel = logging.INFO
    if args.error_format is not None:
        error_formatter = _ErrorFormat[args.error_format]
        default_loglevel = logging.WARNING

    # setup logger again with new configuration
    loglevel = _get_loglevel(args.loglevel, default_loglevel)
    setup_logger(loglevel, pretty=_use_pretty_logging())
    process_output = _is_process_output_requested(
        args.process_output, error_formatter is None
    )

    options = RunOptions(
        require_diagnostics=error_formatter is not None, no_parallel=args.no_parallel
    )
    return _SetupOptions(error_formatter, options, loglevel, process_output)


def _run_target(
    target_name: str,
    runner: Runner,
    base_dir: pathlib.Path,
    args: argparse.Namespace,
    files: Optional[Sequence[pathlib.Path]],
    setup_options: _SetupOptions,
    config: Optional[CliConfig],
) -> None:
    settings_dir: Optional[pathlib.Path] = None
    if config is not None:
        settings_dir = config.settings_dir

    reporter_factory = ReporterFactory(
        pretty=_use_pretty_logging(),
        process_output=setup_options.process_output,
        loglevel=setup_options.loglevel,
    )
    try:
        runner.run(
            target_name,
            base_dir,
            args,
            reporter_factory,
            setup_options.options,
            settings_dir=settings_dir,
            files=files,
        )
    except exceptions.CommandNotFoundError:
        sys.stderr.write(f"target: {target_name} not found\n")
        sys.exit(1)

    error_exit = reporter_factory.has_error()

    if setup_options.error_formatter is None:
        print("\n ** execution summary **")
        print(reporter_factory.format_summary())
        if error_exit:
            sys.stderr.write(f"{target_name} finished with error(s)\n")
            print(reporter_factory.format_error_summary())
    else:
        ret = reporter_factory.format_diagnostic_summary(setup_options.error_formatter)
        if reporter_factory.has_error():
            print(ret)
            print(reporter_factory.format_error_summary(), file=sys.stderr)
        else:
            print("No errors found")

    if error_exit:
        sys.exit(1)


def _start_run(
    base_dir: pathlib.Path,
    runner: Runner,
    config: Optional[CliConfig],
    args: argparse.Namespace,
) -> None:
    target_names = args.targets
    setup_options = _setup_run(base_dir, args, config)
    for target_name in target_names:
        _run_target(
            target_name=target_name,
            runner=runner,
            base_dir=base_dir,
            args=args,
            files=None,
            setup_options=setup_options,
            config=config,
        )


def _start_run_files(
    base_dir: pathlib.Path,
    runner: Runner,
    config: Optional[CliConfig],
    args: argparse.Namespace,
) -> None:
    target_name = args.target
    files = [pathlib.Path(p).resolve() for p in args.files]
    for p in files:
        if not p.exists():
            raise FileNotFoundError(f"{p} does not exist")

    setup_options = _setup_run(base_dir, args, config)
    _run_target(
        target_name=target_name,
        runner=runner,
        base_dir=base_dir,
        args=args,
        files=files,
        setup_options=setup_options,
        config=config,
    )


def _start_generate(
    base_dir: pathlib.Path,
    runner: Runner,
    config: Optional[CliConfig],
    args: argparse.Namespace,
) -> None:
    # NOTE(igarashi): args.export_dir must be resolved by cwd(), not base_dir
    # since it is a cli argument
    settings_dir: pathlib.Path = pathlib.Path(args.export_dir).resolve()
    settings_dir.mkdir(parents=True, exist_ok=True)
    runner.export_settings(base_dir, settings_dir, args)


def _start_list(
    base_dir: pathlib.Path,
    runner: Runner,
    config: Optional[CliConfig],
    args: argparse.Namespace,
) -> None:
    targets = runner.get_targets(args)
    print("available targets:")
    for name, target in targets.items():
        print(f" * {name}")
        for c in target:
            print(f"   - {c}")


def _setup_manifest_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description=CLI_DESCRIPTION, add_help=False)
    parser.add_argument(
        "--config",
        type=str,
        help="Path to a config file",
        default=None,
    )
    parser.add_argument(
        "--version",
        action="store_true",
        help="Show pysen version and exit",
    )
    parser.add_argument(
        "--loglevel",
        type=str,
        help="Set loglevel",
        choices=list(LogLevel.__members__),
    )
    parser.add_argument(
        "--process-output",
        type=str,
        help="Process output control",
        default="auto",
        dest="process_output",
        choices=list(ProcessOutputMode.__members__),
    )
    parser.add_argument(
        "-s",
        action="store_const",
        help="Shortcut for --process-output=show",
        const="show",
        dest="process_output",
    )
    return parser


def _parse_manifest_options() -> Tuple[ManifestBase, Optional[CliConfig], pathlib.Path]:
    # NOTE(igarashi): show detailed help to the user when a configuration is available
    # In this method, we
    # - try to load an available pyproject.toml file
    # - show the help when `--help` is set and pysen cannot load any config
    # - `--help` is handled by `main` parser in `cli()` method in order to show detailed help
    #    when a config is successfully loaded
    parser = _setup_manifest_parser()
    args, unknown = parser.parse_known_args()

    if args.version:
        _show_version()
        sys.exit(0)

    setup_logger(
        _get_loglevel(args.loglevel, logging.INFO), pretty=_use_pretty_logging()
    )

    path: Optional[pathlib.Path] = None
    if args.config is not None:
        path = wrap_path(args.config)
    try:
        pyproject_path = find_pyproject(path)
    except FileNotFoundError as e:
        if "--help" in unknown or "-h" in unknown:
            parser.print_help()
            sys.exit(0)

        sys.stderr.write(f"{e}\n")
        sys.exit(1)

    try:
        manifest = load_manifest(pyproject_path)
        config = cli_config.parse(pyproject_path)
    except exceptions.PysenError as e:
        sys.stderr.write(f"error occurred while loading {pyproject_path}: {e}\n")
        sys.exit(1)

    base_dir = pyproject_path.parent
    return manifest, config, base_dir


def cli() -> None:
    manifest, config, base_dir = _parse_manifest_options()
    runner = Runner(manifest)

    root_parser = _setup_manifest_parser()
    manifest_parser = argparse.ArgumentParser(
        description=CLI_DESCRIPTION,
        parents=[root_parser],
        add_help=False,
    )
    runner.setup_manifest_argparse(
        manifest_parser.add_argument_group("manifest options")
    )
    manifest_args, _ = manifest_parser.parse_known_args()
    targets = runner.get_targets(manifest_args)
    if len(targets) == 0:
        sys.stderr.write(
            "Did not find any targets.\n"
            "A [tool.pysen.lint] section may be missing in your\n"
            "`pyproject.toml` or `pysen.toml`.\n"
            "See the README for details: https://github.com/pfnet/pysen#readme\n"
        )
        sys.exit(1)

    action_parser = argparse.ArgumentParser(parents=[manifest_parser], add_help=True)
    subparsers = action_parser.add_subparsers()
    run_parser = subparsers.add_parser("run", help="run target")
    run_parser.add_argument(
        "targets",
        type=str,
        help="target to run",
        choices=targets,
        nargs="+",
    )
    run_parser.add_argument(
        "--error-format", type=str, choices=_ErrorFormat.keys(), default=None
    )
    run_parser.add_argument("--no-parallel", action="store_true")
    run_parser.set_defaults(func=_start_run)

    run_files_parser = subparsers.add_parser(
        "run_files", help="run target with a specified file"
    )
    run_files_parser.add_argument(
        "target",
        type=str,
        help="target to run",
        choices=targets,
    )
    run_files_parser.add_argument("files", type=str, help="target file", nargs="+")
    run_files_parser.add_argument(
        "--error-format", type=str, choices=_ErrorFormat.keys(), default=None
    )
    run_files_parser.add_argument("--no-parallel", action="store_true")
    run_files_parser.set_defaults(func=_start_run_files)

    generate_parser = subparsers.add_parser("generate", help="generate setting files")
    generate_parser.add_argument(
        "export_dir", type=str, help="target directory to export"
    )
    generate_parser.set_defaults(func=_start_generate)

    list_parser = subparsers.add_parser(
        "list", help="list available targets in manifest"
    )
    list_parser.set_defaults(func=_start_list)

    action_args = action_parser.parse_args()
    # for python 3.6 support, we cannot use add_subparsers(required=True)
    if "func" not in action_args:
        action_parser.print_help()
        sys.exit(1)

    action_args.func(base_dir, runner, config, action_args)



import pathlib

from setuptools import find_packages, setup

BASE_DIR = pathlib.Path(__file__).resolve().parent
exec((BASE_DIR / "pysen/_version.py").read_text())


setup(
    name="pysen",
    version=__version__,  # type: ignore[name-defined]  # NOQA: F821
    packages=find_packages(),
    description=(
        "Python linting made easy. "
        "Also a casual yet honorific way to address individuals "
        "who have entered an organization prior to you."
    ),
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Yuki Igarashi, Toru Ogawa, Ryo Miyajima",
    author_email="igarashi@preferred.jp, ogawa@preferred.jp, ryo@preferred.jp",
    url="https://github.com/pfnet/pysen",
    license="MIT License",
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX",
        "Operating System :: MacOS",
        "Operating System :: Unix",
    ],
    install_requires=[
        "GitPython>=3.0.0,<4.0.0",
        "colorlog>=4.0.0,<5.0.0",
        "dacite>=1.1.0,<2.0.0",
        "dataclasses>=0.6,<1.0;python_version<'3.7'",
        "tomlkit>=0.5.11,<1.0.0",
        "unidiff>=0.6.0,<1.0.0",
    ],
    extras_require={
        "lint": [
            "black>=19.10b0,<=22.10",
            "flake8-bugbear",  # flake8 doesn't have a dependency for bugbear plugin
            "flake8>=3.7,<5",
            "isort>=4.3,<5.2.0",
            "mypy>=0.770,<0.800",
        ],
    },
    package_data={"pysen": ["py.typed"]},
    entry_points={"console_scripts": ["pysen=pysen.cli:cli"]},
)



import contextlib
import logging
import subprocess
import sys
from concurrent.futures import ThreadPoolExecutor
from typing import IO, List, Optional, Sequence, Tuple

from .reporter import Reporter


def _read_stream(stream: Optional[IO[str]], reporter: Reporter, loglevel: int) -> str:
    if stream is None:
        return ""

    ret: List[str] = []
    for line in stream:
        ret.append(line)
        reporter.process_output.log(loglevel, line.rstrip("\n"))

    return "".join(ret)


def add_python_executable(*cmd: str) -> Sequence[str]:
    return [sys.executable, "-m"] + list(cmd)


def run(
    cmd: Sequence[str],
    reporter: Reporter,
    stdout_loglevel: int = logging.INFO,
    stderr_loglevel: int = logging.WARNING,
    encoding: Optional[str] = None,
) -> Tuple[int, str, str]:
    # NOTE: As pysen doesn't configure `sys.stdout` with `errors=ignore` option,
    # it may cause an error when unsupported characters in an environment are
    # going to be printed.
    # As such, `run` method returns strings of printable characters in the environment
    # so that pysen doesn't need to reconfigure `sys.stdout`.
    encoding = encoding or sys.stdout.encoding

    returncode: int = -1
    stdout: str = ""
    stderr: str = ""

    with contextlib.ExitStack() as stack:
        reporter.report_command(" ".join(cmd))
        proc = stack.enter_context(
            subprocess.Popen(
                cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                encoding=encoding,
                universal_newlines=True,
                errors="ignore",
            )
        )
        try:
            pool = stack.enter_context(ThreadPoolExecutor(max_workers=2))
            stdout_task = pool.submit(
                _read_stream, proc.stdout, reporter, stdout_loglevel
            )
            stderr_task = pool.submit(
                _read_stream, proc.stderr, reporter, stderr_loglevel
            )

            proc.wait()

            stdout = stdout_task.result()
            stderr = stderr_task.result()
            returncode = proc.returncode
        except Exception:
            proc.kill()
            raise

    return returncode, stdout, stderr



from typing import Optional

from .path import PathLikeType


class PysenError(Exception):
    pass


class CommandNotFoundError(PysenError):
    pass


class PysenSectionNotFoundError(PysenError):
    pass


class InvalidCommandNameError(PysenError):
    def __init__(self, name: str) -> None:
        super().__init__(f"invalid command name: {name}")


class InvalidConfigurationError(PysenError):
    pass


class InvalidPluginError(PysenError):
    def __init__(self, module_path: Optional[PathLikeType], error: str) -> None:
        super().__init__(f"invalid plugin: {module_path}, {error}")


class InvalidManifestBuilderError(PysenError):
    def __init__(self, path: PathLikeType, error: str) -> None:
        super().__init__(f"invalid manifest builder: {path}, {error}")


class UnexpectedErrorFormat(PysenError):
    pass


class IncompatibleVersionError(PysenError):
    pass


class DistributionNotFound(PysenError):
    pass


class InvalidComponentName(PysenError):
    pass


class RunTargetFileNotSupported(PysenError):
    pass
